#ifndef __MESH_QUADRATURE_C
#define __MESH_QUADRATURE_C

/*
* This code isdistrubuted under GPLv3
* 
* Code Author: Miroslav Stoyanov, Oct 2011
* 
* Copyright (C) 2011  Miroslav Stoyanov
* 
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* Since the GNU General Public License is longer than this entire code, 
* a copy of it can be obtained separately at <http://www.gnu.org/licenses/>
* 
*/

#include "mesh_quadrature.h"

// functions defined over the canonical element
double phi_0( double xi, double et ){ return (1.0 - xi - et); };
double phi_1( double xi, double et ){ return xi; };
double phi_2( double xi, double et ){ return et; };

// derivatives of the canonical functions
double phi_xi_0( double xi, double et ){ return -1.0; };
double phi_xi_1( double xi, double et ){ return  1.0; };
double phi_xi_2( double xi, double et ){ return  0.0; };

double phi_et_0( double xi, double et ){ return -1.0; };
double phi_et_1( double xi, double et ){ return  0.0; };
double phi_et_2( double xi, double et ){ return  1.0; };


void GeometryTriangleSet( double x0, double y0, double x1, double y1, double x2, double y2, struct GeometryTriangle *T ){
        T->x[0] = x0; T->y[0] = y0;
	T->x[1] = x1; T->y[1] = y1;
	T->x[2] = x2; T->y[2] = y2;
        T->xi2x = T->x[1] - T->x[0]; T->et2x = T->x[2] - T->x[0];
        T->xi2y = T->y[1] - T->y[0]; T->et2y = T->y[2] - T->y[0];
        T->det = T->xi2x * T->et2y - T->xi2y * T->et2x;
};

void GeometryQuadratureAdd6PointsTriangle( struct GeometryTriangle T, struct GeometryQuadrature *Q ){
        int i;
	Q->num_points = 6;
	Q->w  = (double*) malloc( 6*sizeof(double) );
	Q->x  = (double*) malloc( 6*sizeof(double) );
	Q->y  = (double*) malloc( 6*sizeof(double) );
	Q->xi = (double*) malloc( 6*sizeof(double) );
	Q->et = (double*) malloc( 6*sizeof(double) );
	
        Q->xi[0] = 0.816847572980459;  Q->et[0] = 0.091576213509771;
        Q->xi[1] = 0.091576213509771;  Q->et[1] = 0.816847572980459;
        Q->xi[2] = 0.091576213509771;  Q->et[2] = 0.091576213509771;
        Q->xi[3] = 0.108103018168070;  Q->et[3] = 0.445948490915965;
        Q->xi[4] = 0.445948490915965;  Q->et[4] = 0.108103018168070;
        Q->xi[5] = 0.445948490915965;  Q->et[5] = 0.445948490915965;
        
        Q->w[0] = 0.5*0.10995174365532*T.det;
        Q->w[1] = 0.5*0.10995174365532*T.det;
        Q->w[2] = 0.5*0.10995174365532*T.det;
        Q->w[3] = 0.5*0.223381589678011*T.det;
        Q->w[4] = 0.5*0.223381589678011*T.det;
        Q->w[5] = 0.5*0.223381589678011*T.det;
        
        for( i=0; i<6; i++ ){
                Q->x[i] = T.xi2x * Q->xi[i] + T.et2x * Q->et[i] + T.x[0];
                Q->y[i] = T.xi2y * Q->xi[i] + T.et2y * Q->et[i] + T.y[0];
        };
};

void GeometryQuadratureAdd3PointsTriangle( struct GeometryTriangle T, struct GeometryQuadrature *Q ){
	int i;
	Q->num_points = 3;
	Q->w  = (double*) malloc( 3*sizeof(double) );
	Q->x  = (double*) malloc( 3*sizeof(double) );
	Q->y  = (double*) malloc( 3*sizeof(double) );
	Q->xi = (double*) malloc( 3*sizeof(double) );
	Q->et = (double*) malloc( 3*sizeof(double) );
	
	Q->xi[0] = 0.5;  Q->et[0] = 0.0;
        Q->xi[1] = 0.0;  Q->et[1] = 0.5;
        Q->xi[2] = 0.5;  Q->et[2] = 0.5;
        
        Q->w[0] = 0.166666666667*T.det;
        Q->w[1] = 0.166666666667*T.det;
        Q->w[2] = 0.166666666667*T.det;
        
        for( i=0; i<3; i++ ){
                Q->x[i] = T.xi2x * Q->xi[i] + T.et2x * Q->et[i] + T.x[0];
                Q->y[i] = T.xi2y * Q->xi[i] + T.et2y * Q->et[i] + T.y[0];
        };
};

void GeometryQuadratureFree( struct GeometryQuadrature *Q ){
	Q->num_points = 0;
	free( Q -> w );
	free( Q -> x );
	free( Q -> y );
	free( Q -> xi );
	free( Q -> et );
};

void GeometryMeshFormUnit( struct GeometryMesh *mesh, double dx, double dy ){
        int i, j;
	int nx =0, ny =0;
	double x =0.0, y =0.0;
	double *ax, *ay; // arrays with the x and y entries
	
	// generating rectangles:
// -------------------------------------------------------------------------------- //
	// generate the list of x entries
        nx = 1.0 / dx + 2; // estimate the nmber of elements and add couple to be safe
	ax = (double*) malloc( nx*sizeof(double) );
	nx = 1; 
	x = 0.0;
	ax[0] = x;
	while ( x < 1.0 ){
		x += dx;
		if ( x > 1.0 ){
			x = 1.0;	
		}else if ( x + dx/2 > 1.0 ){
			x = 1.0;
		};
		ax[nx] = x;
		nx++;
	};
// -------------------------------------------------------------------------------- //
	// generate the list of y entries
	ny = 1.0 / dy + 2; // estimate the nmber of elements and add couple to be safe
	ay = malloc( ny*sizeof(double) );
	ny = 1; 
	y = 0.0;
	ay[0] = y;
	while ( y < 1.0 ){
		y += dy;
		if ( y > 1.0 ){
			y = 1.0;	
		}else if ( y + dy/2 > 1.0 ){
			y = 1.0;
		};
		ay[ny] = y;
		ny++;
	};
// -------------------------------------------------------------------------------- //
	mesh ->elem = (struct GeometryElement*) malloc( 4*(nx-1)*(ny-1)*sizeof( struct GeometryElement ) );
	double cx, cy;
	// setup the triangles and global indexes
	int e;
	for( i=0; i<ny-1; i++ ){
		for( j =0 ; j < nx-1; j++ ){
			cx = 0.5 * ( ax[j] + ax[j+1] );
			cy = 0.5 * ( ay[i] + ay[i+1] );
			
			// First Element
			e = 4*(i * (nx-1) + j);
			// set the triangle coordinates
			GeometryTriangleSet( cx, cy, ax[j], ay[i], ax[j+1], ay[i], &(mesh ->elem[e].T) );
			// set global indexes CL
			mesh ->elem[ e ].igu[0] = i*(2*nx-1)+nx+j;
			mesh ->elem[ e ].igu[1] = i*(2*nx-1)+j;
			mesh ->elem[ e ].igu[2] = i*(2*nx-1)+j+1;
			// set neighbours
			mesh ->elem[ e ].nbrs[0] = (i==0) ? -1 : 4*( (i-1) * (nx-1) + j) +2;
			mesh ->elem[ e ].nbrs[1] = 4*(i * (nx-1) + j) + 1;
			mesh ->elem[ e ].nbrs[2] = 4*(i * (nx-1) + j) + 3;
			
			// Second Element
			e = 4*(i * (nx-1) + j) + 1;
			// set the triangle coordinates
			GeometryTriangleSet( cx, cy, ax[j+1], ay[i], ax[j+1], ay[i+1], &(mesh ->elem[e].T) );
			// set global indexes CL
			mesh ->elem[ e ].igu[0] = i*(2*nx-1)+nx+j;
			mesh ->elem[ e ].igu[1] = i*(2*nx-1)+j+1;
			mesh ->elem[ e ].igu[2] = (i+1)*(2*nx-1)+j+1;
			// set neighbours
			mesh ->elem[ e ].nbrs[0] = (j==nx-2) ? -1 : 4*( i * (nx-1) + j + 1) +3;
			mesh ->elem[ e ].nbrs[1] = 4*( i * (nx-1) + j) +2;
			mesh ->elem[ e ].nbrs[2] = 4*( i * (nx-1) + j);
			
			// Third Element
			e = 4*(i * (nx-1) + j) + 2;
			// set the triangle coordinates
			GeometryTriangleSet( cx, cy, ax[j+1], ay[i+1], ax[j], ay[i+1], &(mesh ->elem[e].T) );
			// set global indexes CL
			mesh ->elem[ e ].igu[0] = i*(2*nx-1)+nx+j;
			mesh ->elem[ e ].igu[1] = (i+1)*(2*nx-1)+j+1;
			mesh ->elem[ e ].igu[2] = (i+1)*(2*nx-1)+j;
			// set neighbours
			mesh ->elem[ e ].nbrs[0] = (i==ny-2) ? -1 : 4*( (i+1) * (nx-1) + j );
			mesh ->elem[ e ].nbrs[1] = 4*( i * (nx-1) + j) +3;
			mesh ->elem[ e ].nbrs[2] = 4*( i * (nx-1) + j) +1;
			
			// Fourth Element
			e = 4*(i * (nx-1) + j) + 3;
			// set the triangle coordinates
			GeometryTriangleSet( cx, cy, ax[j], ay[i+1], ax[j], ay[i], &(mesh ->elem[e].T) );
			// set global indexes CL
			mesh ->elem[ e ].igu[0] = i*(2*nx-1)+nx+j;
			mesh ->elem[ e ].igu[1] = (i+1)*(2*nx-1)+j;
			mesh ->elem[ e ].igu[2] = i*(2*nx-1)+j;
			// set neighbours
			mesh ->elem[ e ].nbrs[0] = (j==0) ? -1 : 4*( i * (nx-1) + j - 1)+1;
			mesh ->elem[ e ].nbrs[1] = 4*( i * (nx-1) + j);
			mesh ->elem[ e ].nbrs[2] = 4*( i * (nx-1) + j) +2;
		};
	};
	mesh ->num_elem = 4*(nx-1)*(ny-1);
	mesh ->num_global_u = (2*nx-1)*(ny-1) + nx;
	
	// set the dirichlet boundary conditoins
	// start by assumig no dirichlet boundary
	for( e=0; e<mesh->num_elem; e++ ){
		for( i=0; i<3; i++ ){
			mesh->elem[ e ].iu[i] = -2;
		};
	};

        // if we have a bottom Dirichlet boundary
        i = 0;
        for( j=0; j < nx-1; j++ ){
                e = 4*(i * (nx-1) + j);
                mesh->elem[ e ].iu[1] = -1;
                mesh->elem[ e ].iu[2] = -1;
                
                e = 4*(i * (nx-1) + j) + 1;
                mesh->elem[ e ].iu[1] = -1;
                
                e = 4*(i * (nx-1) + j) + 3;
                mesh->elem[ e ].iu[2] = -1;
        };
        
        // if we have a bottom Dirichlet boundary
        i = ny-2;
        for( j=0; j < nx-1; j++ ){
                e = 4*(i * (nx-1) + j) + 1;
                mesh->elem[ e ].iu[2] = -1;
                
                e = 4*(i * (nx-1) + j) + 2;
                mesh->elem[ e ].iu[1] = -1;
                mesh->elem[ e ].iu[2] = -1;
                
                e = 4*(i * (nx-1) + j) + 3;
                mesh->elem[ e ].iu[1] = -1;
        };
        
        // if we have a left Dirichlet boundary
        j = 0;
        for( i=0; i<ny-1; i++ ){
                e = 4*(i * (nx-1) + j);
                mesh->elem[ e ].iu[1] = -1;
                
                e = 4*(i * (nx-1) + j) + 2;
                mesh->elem[ e ].iu[2] = -1;
                
                e = 4*(i * (nx-1) + j) + 3;
                mesh->elem[ e ].iu[1] = -1;
                mesh->elem[ e ].iu[2] = -1;
        };

        // if we have a right Dirichlet boundary
        j = nx-2;
        for( i=0; i<ny-1; i++ ){
                e = 4*(i * (nx-1) + j);
                mesh->elem[ e ].iu[2] = -1;
                
                e = 4*(i * (nx-1) + j) + 1;
                mesh->elem[ e ].iu[1] = -1;
                mesh->elem[ e ].iu[2] = -1;
                
                e = 4*(i * (nx-1) + j) + 2;
                mesh->elem[ e ].iu[1] = -1;
        };

	// index the unknown variables
        // there are algorithms that would create a more efficient pattern, this is just something quick that works
	int *indx = (int*) malloc( mesh->num_global_u * sizeof(int) );
	int count = 0;
	for( i=0; i<mesh->num_global_u; i++ ){ indx[i] = -1; };
	
	for( e=0; e<mesh->num_elem; e++ ){
		for( i=0; i<3; i++ ){
			if ( mesh->elem[ e ].iu[i] == -2 ){ // not Dirichlet boundary
				if ( indx[ mesh->elem[ e ].igu[i] ] == -1 ){
					mesh->elem[ e ].iu[i] = count;
					indx[ mesh->elem[ e ].igu[i] ] = count;
					count++;
				}else{
					mesh->elem[ e ].iu[i] = indx[ mesh->elem[ e ].igu[i] ];
				};
			}
		};
	};
	mesh ->num_u = count;
	free( indx );
};

void GeometryMeshFree( struct GeometryMesh *mesh ){
	free( mesh->elem );
	mesh->num_elem = 0;
	mesh->num_u = 0;
	mesh->num_global_u = 0;
};

void GeometryFormLHS( struct GeometryMesh mesh, struct Pmat *L ){
        int e, i, j, k;
        
        struct Pattern p;
        pattern_init( &p, mesh.num_u );
        
        for( e=0; e<mesh.num_elem; e++ ){ // create  the pattern for the sparse matrix
                for( i=0; i<3; i++ ){
                        for( k=0; k<3; k++ ){
                                if ( (mesh.elem[e].iu[i]>-1)&&(mesh.elem[e].iu[k]>-1) ){
                                        pattern_add( &p, mesh.elem[e].iu[k], mesh.elem[e].iu[i] );
                                };
                        };
                };
        };
        
        pmat_create( L, p ); // create the matrix with the pattern and zero all values
        
        pattern_del( &p ); // release the memory allocated for the pattern
        
        double (*phi_xi[3])(double, double) = { &phi_xi_0, &phi_xi_1, &phi_xi_2 }; // set the three test functions into an array of functions
        double (*phi_et[3])(double, double) = { &phi_et_0, &phi_et_1, &phi_et_2 }; // set the three test functions into an array of functions
        
        struct GeometryQuadrature Q;
        double I;
        for( e=0; e<mesh.num_elem; e++ ){
                for( i=0; i<3; i++ ){
                        for( k=0; k<3; k++ ){
                                if ( (mesh.elem[e].iu[i]>-1)&&(mesh.elem[e].iu[k]>-1) ){
                                        GeometryQuadratureAdd6PointsTriangle( mesh.elem[e].T, &Q );
                                        I = 0.0;
                                        for( j=0; j<Q.num_points; j++ ){
                                                I += (Q.w[j] / (mesh.elem[e].T.det * mesh.elem[e].T.det)) * (
							(mesh.elem[e].T.et2y*phi_xi[i](Q.xi[j],Q.et[j]) 
									- mesh.elem[e].T.xi2y*phi_et[i](Q.xi[j],Q.et[j]) )*
							(mesh.elem[e].T.et2y*phi_xi[k](Q.xi[j],Q.et[j]) 
									- mesh.elem[e].T.xi2y*phi_et[k](Q.xi[j],Q.et[j]) )+
							(-mesh.elem[e].T.et2x*phi_xi[i](Q.xi[j],Q.et[j]) 
									+ mesh.elem[e].T.xi2x*phi_et[i](Q.xi[j],Q.et[j]) )*
							(-mesh.elem[e].T.et2x*phi_xi[k](Q.xi[j],Q.et[j]) 
									+ mesh.elem[e].T.xi2x*phi_et[k](Q.xi[j],Q.et[j]) ) );
                                        };
                                        pmat_addval( L, mesh.elem[e].iu[k], mesh.elem[e].iu[i], -I );
                                        GeometryQuadratureFree( &Q );
                                };
                        };
                };
        };
};

void GeometryFromRHS( struct GeometryMesh mesh, double *R, double (*f)(double,double), double (*boundary)(double,double) ){
	int e, i, j, k;
	double (*phi[3])(double, double) = { &phi_0, &phi_1, &phi_2 };
	double (*phi_xi[3])(double, double) = { &phi_xi_0, &phi_xi_1, &phi_xi_2 }; // set the three test functions into an array of functions
        double (*phi_et[3])(double, double) = { &phi_et_0, &phi_et_1, &phi_et_2 }; // set the three test functions into an array of functions
	struct GeometryQuadrature Q;
        double I;
        
        for( i=0; i<mesh.num_u; i++ ){
		R[i] = 0.0;
	};
	
	// do the integral f * phi
	for( e=0; e<mesh.num_elem; e++ ){
		for( k=0; k<3; k++ ){
			if ( mesh.elem[e].iu[k] > -1 ){
				GeometryQuadratureAdd6PointsTriangle( mesh.elem[e].T, &Q );
				I = 0.0;
                                for( j=0; j<Q.num_points; j++ ){
					I += Q.w[j] * phi[k](Q.xi[j],Q.et[j]) * f(Q.x[j],Q.y[j]);
				};
				R[mesh.elem[e].iu[k]] += I;
				GeometryQuadratureFree( &Q );
			};
		};
	};
	
	// do the boundary conditions
	for( e=0; e<mesh.num_elem; e++ ){
                for( i=0; i<3; i++ ){
                        for( k=0; k<3; k++ ){
                                if ( (mesh.elem[e].iu[i]==-1)&&(mesh.elem[e].iu[k]>-1) ){
                                        GeometryQuadratureAdd6PointsTriangle( mesh.elem[e].T, &Q );
                                        I = 0.0;
                                        for( j=0; j<Q.num_points; j++ ){
                                                I += (Q.w[j] / (mesh.elem[e].T.det * mesh.elem[e].T.det)) * (
							(mesh.elem[e].T.et2y*phi_xi[i](Q.xi[j],Q.et[j]) 
									- mesh.elem[e].T.xi2y*phi_et[i](Q.xi[j],Q.et[j]) )*
							(mesh.elem[e].T.et2y*phi_xi[k](Q.xi[j],Q.et[j]) 
									- mesh.elem[e].T.xi2y*phi_et[k](Q.xi[j],Q.et[j]) )+
							(-mesh.elem[e].T.et2x*phi_xi[i](Q.xi[j],Q.et[j]) 
									+ mesh.elem[e].T.xi2x*phi_et[i](Q.xi[j],Q.et[j]) )*
							(-mesh.elem[e].T.et2x*phi_xi[k](Q.xi[j],Q.et[j]) 
									+ mesh.elem[e].T.xi2x*phi_et[k](Q.xi[j],Q.et[j]) ) );
                                        };
                                        R[mesh.elem[e].iu[k]] += I * boundary( mesh.elem[e].T.x[i], mesh.elem[e].T.y[i] );
                                        GeometryQuadratureFree( &Q );
                                };
                        };
                };
        };
};

void GeometryLoadSolution( struct GeometryMesh *mesh, double *u, double (*boundary)(double,double) ){
	int e, i;
	for( e=0; e<mesh->num_elem; e++ ){
		for( i=0; i<3; i++ ){
			mesh->elem[e].u[i] = ( mesh->elem[e].iu[i]>-1 ) ? u[mesh->elem[e].iu[i]] : boundary( mesh->elem[e].T.x[i], mesh->elem[e].T.y[i] );
		};
	};
};

double GeometryComputerErrorL2( struct GeometryMesh mesh, double (*u_exact)(double,double) ){
	int e, i, j;
	double (*phi[3])(double, double) = { &phi_0, &phi_1, &phi_2 };
	struct GeometryQuadrature Q;
	double err = 0.0;
        double I;
        double u, ue;
        
        for( e=0; e<mesh.num_elem; e++ ){
		GeometryQuadratureAdd6PointsTriangle( mesh.elem[e].T, &Q );
		for( j=0; j<Q.num_points; j++ ){
			u = 0.0;
			for( i=0; i<3; i++ ){ // get the value of the solution at the quadrature point
				u += mesh.elem[e].u[i] * phi[i]( Q.xi[j], Q.et[j] );
			};
			ue = u_exact( Q.x[j], Q.y[j] );
			err += Q.w[j] * ( u - ue ) * ( u - ue );
		};
		GeometryQuadratureFree( &Q );
	};
	return sqrt( err );
};

void GeometrySaveSolution( struct GeometryMesh mesh, char *filename ){
	FILE * fout = fopen( filename, "w");
	fprintf(fout,"%d\n",mesh.num_elem);
	int e;
	for( e=0; e<mesh.num_elem; e++ ){
		fprintf(fout,"%f %f %f %f %f %f %f %f %f\n",mesh.elem[e].T.x[0],mesh.elem[e].T.y[0],
			mesh.elem[e].T.x[1],mesh.elem[e].T.y[1],
			mesh.elem[e].T.x[2],mesh.elem[e].T.y[2],
			mesh.elem[e].u[0],mesh.elem[e].u[1],mesh.elem[e].u[2]);
	};
	fclose( fout );
};

#endif
